# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, AsyncGenerator, Awaitable, Sequence, Type

import asyncpg
from asyncpg.cursor import CursorFactory

from .connection import Connection, Pool
from .exceptions import NoMigrationsToCreate
from .indexes import Index
from .migrations import describe
from .migrations.applied_migration import AppliedMigration
from .migrations.apply_migration import apply_migration
from .migrations.create_migration import create_next_migration
from .migrations.migration import Migration
from .model import Model
from .utils.lazy_list import LazyList


class Database:
    """Base database class. You must subclass this, as
    all tables and indexes are stored as classvars.

    Example:
    ```
    class MyDatabase(Database):
        users = User
        games = Game
        players = Players

        indexes = [SomeIndex, SomeOtherIndex]
    ```
    """

    _migrations = AppliedMigration
    """Internal table to track applied migrations."""

    indexes: Sequence[Index] | None = None
    """A list of indexes for the database."""

    def __init__(self, migrations_folder: Path | str) -> None:
        """Initialize the database.

        Args:
            migrations_folder (pathlib.Path): The folder in which migrations
            are/will be stored.
        """

        self._migrations_folder = (
            Path(migrations_folder)
            if isinstance(migrations_folder, str)
            else migrations_folder
        )
        self._all_models: list[Type[Model]] = []

        for attr_name in dir(self):
            try:
                attr = getattr(self, attr_name)
            except AttributeError:  # pragma: no cover
                continue

            if not isinstance(attr, type):
                continue
            if not issubclass(attr, Model):
                continue

            attr.database = self
            attr.tablename = attr_name
            self._all_models.append(attr)

        self.pool: Pool | None = None

    # migration functions
    def describe(self) -> describe.Describe:
        """Return the description of the database.

        Returns:
            describe.Describe: The description.
        """

        return describe.Describe(
            tables=[m._describe() for m in self._all_models],
            indexes=[i._describe() for i in self.indexes or []],
        )

    def load_all_migrations(self) -> list[Migration]:
        """Load all migrations and return them.

        Returns:
            list[Migration]: The migrations.
        """

        return Migration._load_all_migrations(self._migrations_folder)

    def load_last_migration(self) -> Migration | None:
        """Loads and returns the most recent migration, if any.

        Returns:
            Migration | None: The migration.
        """

        return Migration._load_last_migration(self._migrations_folder)

    def load_migration_from_id(self, migration_id: int) -> Migration:
        """Loads a migration from its id.

        Args:
            migration_id (int): The id of the migration.

        Returns:
            Migration: The migration.
        """

        return Migration._from_path(
            Migration._path_from_id(migration_id, self._migrations_folder)
        )

    def must_create_migrations(self) -> bool:
        """Whether or not you need to call Database.create_migrations()

        Returns:
            bool
        """

        sql = self._create_next_migration()
        if sql is None:
            return False
        return True

    def create_migrations(self, allow_empty: bool = False) -> Migration:
        """Create migrations.

        Args:
            allow_empty (bool): Whether to allow the migration to be empty
            (useful for creating custom migrations).

        Raises:
            NoMigrationsToCreate: Migrations do not creating.

        Returns:
            Migration: The created migration.
        """

        if (not self.must_create_migrations()) and not allow_empty:
            raise NoMigrationsToCreate
        sql = self._create_next_migration()
        sql = sql or ""
        return Migration._create_migration(
            self.describe(), sql, self._migrations_folder
        )

    async def load_unapplied_migrations(self) -> list[Migration]:
        """Returns a list of migrations that have not been applied on the
        current database.

        Returns:
            list[Migration]: The unapplied migrations.
        """

        try:
            applied = [
                m.id_ for m in await self._migrations.fetch_query().fetchmany()
            ]
        except asyncpg.UndefinedTableError:
            applied = []
        return [
            m
            for m in self.load_all_migrations()
            if m.migration_id not in applied
        ]

    async def must_apply_migrations(self) -> bool:
        """Whether or not there are migrations that need to be applied.

        Returns:
            bool
        """

        return len(await self.load_unapplied_migrations()) > 0

    async def apply_migrations(self) -> None:
        """Applies all migrations that need to be applied."""

        for m in await self.load_unapplied_migrations():
            await self._apply_migration(m)

    # database functions
    async def connect(self, **connect_kwargs) -> None:
        """Connect to a database. Any kwargs that can be passed to
        asyncpg.create_pool() can be used here.
        """

        self.pool = Pool(await asyncpg.create_pool(**connect_kwargs))

    async def cleanup(self, timeout: float = 30) -> None:
        """Close the connection.

        Args:
            timeout (float): The maximum time pool.close() has. Defaults to 30.

        Raises:
            TimeoutError: The operation timed out.
        """

        if self.pool is not None:
            await asyncio.wait_for(self.pool.close(), timeout=timeout)

    async def execute(self, query: str, params: list[Any]) -> None:
        """Execute SQL within a transaction."""

        assert self.pool is not None
        async with self.pool.acquire() as con:
            async with con.transaction():
                await con.execute(query, params)

    async def fetchrow(
        self, query: str, params: list[Any]
    ) -> dict[str, Any] | None:
        """Fetch the first matching row.

        Returns:
            dict | None: The row, if any.
        """

        assert self.pool is not None
        async with self.pool.acquire() as con:
            async with con.transaction():
                return await con.fetchrow(query, params)

    async def fetchmany(
        self, query: str, params: list[Any]
    ) -> LazyList[asyncpg.Record, dict[str, Any]]:
        """Fetch all matching rows.

        Returns:
            LazyList[asyncpg.Record, dict]: All matching rows.
        """

        assert self.pool is not None
        async with self.pool.acquire() as con:
            async with con.transaction():
                return await con.fetchmany(query, params)

    async def fetchval(self, query: str, params: list[Any]) -> Any:
        """Fetch a single value."""

        assert self.pool is not None
        async with self.pool.acquire() as con:
            async with con.transaction():
                return await con.fetchval(query, params)

    @asynccontextmanager
    async def cursor(
        self, query: str, params: list[Any], con: Connection | None = None
    ) -> AsyncGenerator[CursorFactory, None]:
        """Yields a CursorFactory.

        Usage:
        ```
        async with db.cursor(query, args) as cursor:
            async for res in cursor:
                print(res)
        ```
        """

        if con:
            yield con.cursor(query, params)

        else:
            assert self.pool is not None
            async with self.pool.acquire() as con:
                async with con.transaction():
                    yield con.cursor(query, params)

    def _create_next_migration(self) -> str | None:
        return create_next_migration(self.describe(), self._migrations_folder)

    def _apply_migration(self, migration: Migration) -> Awaitable[None]:
        return apply_migration(migration, self)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Sequence, Type

from .exceptions import BadArgument
from .field import BaseField
from .migrations.describe import DescribeIndex
from .sql.sql import Block, join, raw, wrap

if TYPE_CHECKING:  # pragma: no cover
    from .model import Model


@dataclass
class _IndexType:
    name: str
    multi: bool = False
    unique: bool = False


class IndexType(Enum):
    """An enum containing every supported index type."""

    BTREE = _IndexType("BTREE", multi=True, unique=True)
    HASH = _IndexType("HASH")
    GIST = _IndexType("GIST", multi=True)
    SPGIST = _IndexType("SPGIST")
    GIN = _IndexType("GIN", multi=True)
    BRIN = _IndexType("BRIN", multi=True)


class Index:
    """Represents an index for a table.

    Must be added to Database.indexes before the database is initialized. For
    example:
    ```
    class MyDatabase(Database):
        ...

        indexes = [Index(...), Index(...), ...]
    ```

    Args:
        table (Type[Model]): The table this index is for.
        fields (Sequence[BaseField): A list of fields that this index is for.
        type_ (IndexType, optional): The index type. Defaults to
        IndexType.BTREE.
        unique (bool, optional): Whether or not the index should be unique
        (BTREE only). Defaults to False.

    Raises:
        BadArgument: Bad arguments were passed to Index.
    """

    def __init__(
        self,
        table: Type[Model],
        fields: Sequence[BaseField | Block | str] | BaseField | Block | str,
        type_: IndexType = IndexType.BTREE,
        unique: bool = False,
    ) -> None:
        self.type_: _IndexType = type_.value
        if not isinstance(fields, Sequence):
            fields = [fields]
        self.fields = [raw(f) if isinstance(f, str) else f for f in fields]
        self.table = table
        self.unique = unique

        if len(fields) == 0:
            raise BadArgument("Must specify at least one field for index.")

        if (not self.type_.multi) and len(fields) > 1:
            raise BadArgument(f"{type_.name} indexes only support one column.")

        if (not self.type_.unique) and unique:
            raise BadArgument(
                f"{type_.name} indexes do not support uniqueness."
            )

    def get_name(self) -> str:
        """Create a name based on the type, table, and fields of the index.

        Returns:
            str: The name.
        """

        fields = [f.name for f in self.fields if isinstance(f, BaseField)]
        return "_{type_}_index_{table}__{cols}".format(
            type_=self.type_.name,
            table=self.table.tablename,
            cols="_".join(fields),
        ).lower()

    def _creation_sql(self) -> Block:
        names = [
            wrap(f) if isinstance(f, Block) else wrap(raw(f.name))
            for f in self.fields
        ]

        return Block(
            (raw("UNIQUE INDEX") if self.unique else raw("INDEX")),
            raw(self.get_name()),
            raw("ON"),
            raw(self.table.tablename),
            raw("USING"),
            raw(self.type_.name),
            raw("("),
            join(raw(","), *names),
            raw(")"),
        )

    def _describe(self) -> DescribeIndex:
        return DescribeIndex(
            name=self.get_name(),
            raw_sql=self._creation_sql().render_no_params(),
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from .base_type import SqlType


class VarChar(SqlType[str]):
    """Variable length string with limit.

    Args:
        max_length (int, optional): The maximum length of the string. If
        None, any length will be accepted. Defaults to None.

    https://www.postgresql.org/docs/14/datatype-character.html
    """

    def __init__(self, max_length: int | None = None) -> None:
        self._max_length = max_length
        self._sql = "VARCHAR"
        if max_length is not None:
            self._sql += f"({max_length})"

    @property
    def max_length(self) -> int | None:
        """The maximum length of the VarChar type.

        Returns:
            int | None
        """

        return self._max_length


class Char(SqlType[str]):
    """Fixed length string. If a passed string is too short, it will be
    padded with spaces.

    Args:
        max_length (int, optional): The length of the string. If None,
        postgres will use 1. Defaults to None.

    https://www.postgresql.org/docs/14/datatype-character.html
    """

    def __init__(self, length: int | None = None) -> None:
        self._length = length
        self._sql = "CHAR"
        if length is not None:
            self._sql += f"({length})"

    @property
    def length(self) -> int | None:
        """The length of the string.

        Returns:
            int | None
        """

        return self._length


class Text(SqlType[str]):
    """Variable unlimited length string.

    https://www.postgresql.org/docs/14/datatype-character.html
    """

    _sql = "TEXT"



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

from pydantic import BaseModel


class DescribeField(BaseModel):
    """Field description."""

    name: str
    """The name of the field."""
    type_: str
    """The raw SQL name of the type (INTEGER, VARCHAR, etc.)."""
    not_null: bool
    """Whether or not the field should be marked NOT NULL."""


class DescribeConstraint(BaseModel):
    """Constraint description."""

    name: str
    """The name of the constraint."""
    raw_sql: str
    """The raw SQL of the constraint."""


class DescribeIndex(BaseModel):
    """Index description."""

    name: str
    """The name of the index."""
    raw_sql: str
    """The raw SQL of the index."""


class DescribeTable(BaseModel):
    """Table description."""

    name: str
    """The name of the table."""
    fields: List[DescribeField]
    """List of field descriptions."""
    fk_constraints: List[DescribeConstraint]
    """List of ForeignKey descriptions."""
    pk_constraint: DescribeConstraint
    """List of PrimaryKey descriptions."""
    unique_constraints: List[DescribeConstraint]
    """List of Unique descriptions."""
    check_constraints: List[DescribeConstraint]
    """List of Check descriptions."""
    exclude_constraints: List[DescribeConstraint]
    """List of Exclude descriptions."""

    @property
    def constraints(self) -> List[DescribeConstraint]:
        """List of all constraints (FK, PK, unique, check, and exclude).

        Returns:
            List[DescribeConstraint]: List of constraints.
        """

        return (
            self.fk_constraints
            + [self.pk_constraint]
            + self.unique_constraints
            + self.check_constraints
            + self.exclude_constraints
        )


class Describe(BaseModel):
    """Database description."""

    tables: List[DescribeTable]
    """List of table descriptions."""
    indexes: List[DescribeIndex]
    """List of index descriptions."""



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Type,
    TypeVar,
    overload,
)

from .converter import Converter
from .exceptions import BadArgument, InvalidFieldValue, UndefinedFieldValue
from .migrations.describe import DescribeField
from .sql.sql import Comparable, raw
from .undefined import UNDEF

if TYPE_CHECKING:  # pragma: no cover
    from .model import Model
    from .sql.sql import Block
    from .types.base_type import SqlType


_SELF = TypeVar("_SELF", bound="BaseField", covariant=True)
_T = TypeVar("_T")
_C = TypeVar("_C")
_F = TypeVar("_F", bound="SqlType")
VALIDATOR = Callable[[_C], bool]


class BaseField(Comparable, Generic[_F, _T, _C]):
    """Represents a field for a table."""

    name: str  # populated by Database
    """The name of the field, populated when Database is initialized."""
    model: Type[Model]  # populated by Database
    """The model this field belongs to, populated when Database is
    initialized."""

    def __init__(
        self,
        sql_type: _F,
        *,
        default: _T | UNDEF = UNDEF.UNDEF,
        default_factory: Callable[[], _T] | None = None,
        not_null: bool = False,
        use_repr: bool = True,
    ) -> None:
        self.sql_type = sql_type

        if (default is not UNDEF.UNDEF) and (default_factory is not None):
            raise BadArgument("Cannot specify default and default_factory.")

        self._default = default
        self._default_factory = default_factory

        self.not_null = not_null

        self.use_repr = use_repr

        self._validators: list[VALIDATOR[_C]] = []

    @overload
    def __get__(self: _SELF, inst: None, cls: Type[Model]) -> _SELF:
        ...

    @overload
    def __get__(self, inst: Model, cls: Type[Model]) -> _C:
        ...

    def __get__(
        self: _SELF, inst: Model | None, cls: Type[Model]
    ) -> _SELF | _C:
        raise NotImplementedError

    def __set__(self, inst: Model, value: _C):
        raise NotImplementedError

    @property
    def full_name(self) -> str:
        """The full name of the field (`tablename.fieldname`)."""

        return f"{self.model.tablename}.{self.name}"

    def add_validator(self: _SELF, validator: VALIDATOR[_C]) -> _SELF:
        """Add a validator to the value of this field."""

        self._validators.append(validator)
        return self

    def _validate(self, value: _C) -> None:
        for v in self._validators:
            try:
                if not v(value):
                    raise InvalidFieldValue(
                        f"Validator on {self.name} failed for {value!r}"
                    )
            except InvalidFieldValue:
                raise

    def _describe(self) -> DescribeField:
        return DescribeField(
            name=self.name,
            type_=self.sql_type._sql,
            not_null=self.not_null,
        )

    def _get_default(self) -> _T | UNDEF:
        if self._default is not UNDEF.UNDEF:
            return self._default
        if self._default_factory is not None:
            return self._default_factory()
        return UNDEF.UNDEF

    def _get_block(self) -> Block:
        return raw(self.full_name)

    def _copy_kwargs(self) -> dict[str, Any]:
        return dict(
            sql_type=self.sql_type,
            default=self._default,
            default_factory=self._default_factory,
            not_null=self.not_null,
            use_repr=self.use_repr,
        )


class Field(BaseField[_F, _T, _T]):
    if not TYPE_CHECKING:

        def __get__(self, inst, cls):
            if not inst:
                return self
            if self.name not in inst._raw_values:
                raise UndefinedFieldValue(self)
            return inst._raw_values[self.name]

        def __set__(self, inst, value):
            self._validate(value)
            inst._raw_values[self.name] = value
            inst._changed_fields.add(self.name)

    def with_converter(
        self, converter: Converter[_T, _C] | Type[Converter[_T, _C]]
    ) -> ConverterField[_F, _T, _C]:
        """Add a converter to the field.

        Args:
            converter (Converter | Type[Converter]): The converter.

        Returns:
            ConverterField: The new field with the converter.
        """

        if isinstance(converter, type) and issubclass(converter, Converter):
            converter = converter()
        f: ConverterField[_F, _T, _C] = ConverterField(
            **self._copy_kwargs(), converter=converter
        )
        if hasattr(self, "name"):
            f.name = self.name
        if hasattr(self, "model"):
            f.model = self.model
        return f


class ConverterField(BaseField[_F, _T, _C]):
    def __init__(self, *args, **kwargs) -> None:
        self.converter: Converter[_T, _C] = kwargs.pop("converter")
        super().__init__(*args, **kwargs)

    if not TYPE_CHECKING:

        def __get__(
            self, inst: Model | None, cls: Type[Model]
        ) -> ConverterField | Any:
            if not inst:
                return self
            if self.name not in inst._raw_values:
                raise UndefinedFieldValue(self)
            return self.converter.from_stored(inst._raw_values[self.name])

        def __set__(self, inst, value):
            self._validate(value)
            inst._raw_values[self.name] = self.converter.to_stored(value)
            inst._changed_fields.add(self.name)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, Type

if TYPE_CHECKING:  # pragma: no cover
    from .field import BaseField
    from .model import Model


class ApgormBaseException(Exception):
    """The base clase for all exceptions in apgorm."""


# migration-side exceptions
class MigrationException(ApgormBaseException):
    """Base class for all exceptions related to migrations."""


class NoMigrationsToCreate(MigrationException):
    """The migration for the given id was not found."""

    def __init__(self) -> None:
        super().__init__("There are no migrations to create.")


class MigrationAlreadyApplied(MigrationException):
    """The migration has already been applied."""

    def __init__(self, path: str) -> None:
        super().__init__(f"The migration at {path} has already been applied.")


# apgorm-side exceptions
class ApgormException(ApgormBaseException):
    """Base class for all exceptions related to the code of apgorm."""


class UndefinedFieldValue(ApgormException):
    """Raised if you try to get the value for a field that is undefined.

    Usually means that the model has not been created."""

    def __init__(self, field: BaseField) -> None:
        self.field = field

        super().__init__(
            f"The field {field.full_name} is undefined. "
            "This usually means that the model has not been "
            "created."
        )


class InvalidFieldValue(ApgormException):
    """The field value failed the validator check."""

    def __init__(self, message: str) -> None:
        self.message = message

        super().__init__(message)


class SpecifiedPrimaryKey(ApgormException):
    """You tried to create a primary key constraint by using PrimaryKey
    instead of Model.primary_key."""

    def __init__(self, cls: str, fields: Sequence[str]) -> None:
        super().__init__(
            f"You tried to specify a primary key on {cls} by using "
            f"the PrimaryKey constraint. Please use {cls}.primary_key "
            "instead:\nprimary_key = ({},)".format(", ".join(fields))
        )


class BadArgument(ApgormException):
    """Bad arguments were passed."""

    def __init__(self, message: str) -> None:
        self.message = message

        super().__init__(message)


# postgres-side exceptions
class SqlException(ApgormBaseException):
    """Base class for all exceptions related to SQL."""


class ModelNotFound(SqlException):
    """A model for the given parameters was not found."""

    def __init__(self, model: Type[Model], values: dict[str, Any]) -> None:
        self.model = model
        self.values = values

        super().__init__(
            "No Model was found for the following parameters:\n"
            " - "
            + ("\n - ".join([f"{k!r} = {v!r}" for k, v in values.items()]))
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from apgorm.sql.sql import SQL, Block, raw


def add_table(tablename: Block) -> Block:
    return Block(raw("CREATE TABLE"), tablename, raw("()"))


def drop_table(tablename: Block) -> Block:
    return Block(raw("DROP TABLE"), tablename)


def add_index(raw_sql: str) -> Block:
    return Block(raw("CREATE"), raw(raw_sql))


def drop_index(name: Block) -> Block:
    return Block(raw("DROP INDEX"), name)


def _alter_table(tablename: Block, sql: SQL) -> Block:
    return Block(raw("ALTER TABLE"), tablename, sql)


def add_constraint(
    tablename: Block,
    constraint_raw_sql: str,
) -> Block:
    return _alter_table(tablename, Block(raw("ADD"), raw(constraint_raw_sql)))


def drop_constraint(tablename: Block, constraint_name: Block) -> Block:
    return _alter_table(
        tablename, Block(raw("DROP CONSTRAINT"), constraint_name)
    )


def add_field(tablename: Block, fieldname: Block, type_: Block) -> Block:
    return _alter_table(
        tablename,
        Block(
            raw("ADD COLUMN"),
            fieldname,
            type_,
        ),
    )


def drop_field(tablename: Block, fieldname: Block) -> Block:
    return _alter_table(tablename, Block(raw("DROP COLUMN"), fieldname))


def _alter_field(tablename: Block, fieldname: Block, sql: SQL) -> Block:
    return _alter_table(
        tablename,
        Block(
            raw("ALTER COLUMN"),
            fieldname,
            sql,
        ),
    )


def set_field_not_null(
    tablename: Block, fieldname: Block, not_null: bool
) -> Block:
    return _alter_field(
        tablename,
        fieldname,
        Block(
            raw("SET NOT NULL") if not_null else raw("DROP NOT NULL"),
        ),
    )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

from apgorm.sql.sql import Block, raw

from .constraint import Constraint

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.types.boolean import Bool


class Check(Constraint):
    def __init__(self, check: Block[Bool] | str) -> None:
        """Specify a check constraint for a table.

        Args:
            check (Block[Bool]): The raw SQL for the check constraint.
        """

        self.check = raw(check) if isinstance(check, str) else check

    def _creation_sql(self) -> Block:
        return Block(
            raw("CONSTRAINT"),
            raw(self.name),
            raw("CHECK"),
            Block(self.check, wrap=True),
            wrap=True,
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

import asyncpg

from apgorm.exceptions import MigrationAlreadyApplied, ModelNotFound

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.database import Database

    from .migration import Migration


async def apply_migration(migration: Migration, db: Database) -> None:
    try:
        await db._migrations.fetch(id_=migration.migration_id)
    except (asyncpg.exceptions.UndefinedTableError, ModelNotFound):
        pass
    else:
        raise MigrationAlreadyApplied(str(migration.path))

    assert db.pool is not None
    async with db.pool.acquire() as con:
        async with con.transaction():
            await con.execute(migration.migrations, [])
            await db._migrations(id_=migration.migration_id).create(con=con)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apgorm',
 'apgorm.constraints',
 'apgorm.migrations',
 'apgorm.sql',
 'apgorm.sql.generators',
 'apgorm.types',
 'apgorm.utils']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg>=0.25.0,<0.26.0', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'apgorm',
    'version': '1.0.0b1',
    'description': 'A fully type-checked asynchronous ORM wrapped around asyncpg.',
    'long_description': '# apgorm\n[![pytest](https://github.com/TrigonDev/apgorm/actions/workflows/pytest.yml/badge.svg)](https://github.com/TrigonDev/apgorm/actions/workflows/pytest.yml)\n[![pypi](https://github.com/TrigonDev/apgorm/actions/workflows/pypi.yml/badge.svg)](https://pypi.org/project/apgorm)\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/TrigonDev/apgorm/main.svg)](https://results.pre-commit.ci/latest/github/TrigonDev/apgorm/main)\n[![codecov](https://codecov.io/gh/TrigonDev/apgorm/branch/main/graph/badge.svg?token=LEY276K4NS)](https://codecov.io/gh/TrigonDev/apgorm)\n\n[Documentation](https://github.com/trigondev/apgorm/wiki) | [CONTRIBUTING.md](https://github.com/trigondev/.github/tree/main/CONTRIBUTING.md)\n\nAn asynchronous ORM wrapped around asyncpg. Examples can be found under `examples/`. Run examples with `python -m examples.<example_name>` (`python -m examples.basic`).\n\nPlease note that this library is not for those learning SQL or Postgres. Although the basic usage of apgorm is straightforward, you will run into problems, especially with migrations, if you don\'t understand regular SQL well.\n\nIf you need support, you can contact me `Circuit#3397` after joining [this server](https://discord.gg/dGAzZDaTS9). I don\'t accept friend requests.\n\n## Features\n - Fairly straightforward and easy-to-use.\n - Support for basic migrations.\n - Protects against SQL-injection.\n - Python-side converters and validators.\n - Decent many-to-many support.\n - Fully type-checked.\n - Tested.\n\n## Limitations\n - Limited column namespace. For example, you cannot have a column named `tablename` since that is used to store the name of the model.\n - Only supports PostgreSQL with asyncpg.\n - Migrations don\'t natively support field/table renaming, but you can still write your own migration with raw SQL.\n\n## Basic Usage\nDefining a model and database:\n```py\nclass User(apgorm.Model):\n    username = VarChar(32).field()\n    email = VarChar().nullablefield()\n    \n    primary_key = (username,)\n    \nclass Database(apgorm.Database):\n    users = User\n```\n\nIntializing the database:\n```py\ndb = Database(migrations_folder="path/to/migrations")\nawait db.connect(database="database name")\n```\n\nCreating & Applying migrations:\n```py\nif db.must_create_migrations():\n    db.create_migrations()\nif await db.must_apply_migrations():\n    await db.apply_migrations()\n```\n\nBasic create, fetch, update, and delete:\n```py\nuser = await User(username="Circuit").create()\nprint("Created user", user)\n\nassert user == await User.fetch(username="Circuit")\n\nuser.email.v = "email@example.com"\nawait user.save()\n\nawait user.delete()\n```\n',
    'author': 'Circuit',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/TrigonDev/apgorm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import json
from pathlib import Path

from .describe import Describe


class Migration:
    """Represents a single migrations."""

    def __init__(
        self, describe: Describe, migrations: str, path: Path
    ) -> None:
        self.describe = describe
        """The Database.describe() at the time this migration was created."""
        self.migrations = migrations
        """The raw SQL used for applying the migration."""
        self.path = path
        """The path at which the migration can be found."""

    @property
    def migration_id(self) -> int:
        """The id of the migration.

        Returns:
            int: The migration id.
        """

        return self._id_from_path(self.path)

    @classmethod
    def _from_path(cls, path: Path) -> Migration:
        with (path / "describe.json").open("r") as f:
            describe = Describe(**json.loads(f.read()))

        with (path / "migrations.sql").open("r") as f:
            sql = f.read()

        return cls(describe, sql, path)

    @staticmethod
    def _path_from_id(migration_id: int, folder: Path) -> Path:
        return folder / str(migration_id)

    @staticmethod
    def _id_from_path(path: Path) -> int:
        return int(path.name)

    @classmethod
    def _load_all_migrations(cls, folder: Path) -> list[Migration]:
        migrations: list[Migration] = []
        for p in folder.glob("*"):
            if p.is_file():
                continue

            try:
                migrations.append(cls._from_path(p))
            except FileNotFoundError:
                continue

        return migrations

    @classmethod
    def _load_last_migration(cls, folder: Path) -> Migration | None:
        migrations = cls._load_all_migrations(folder)
        if len(migrations) == 0:
            return None
        migrations.sort(key=lambda m: m.migration_id)
        return migrations[-1]

    @classmethod
    def _create_migration(
        cls, describe: Describe, migrations: str, folder: Path
    ) -> Migration:
        last_migration = cls._load_last_migration(folder)
        if last_migration:
            next_id = last_migration.migration_id + 1
        else:
            next_id = 0
        next_path = folder / str(next_id)
        next_path.mkdir(parents=True, exist_ok=False)
        with (next_path / "describe.json").open("w+") as f:
            f.write(json.dumps(describe.dict(), indent=4))
        with (next_path / "migrations.sql").open("w+") as f:
            f.write(migrations)

        return cls(describe, migrations, next_path)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Migration):
            raise TypeError(f"Unsupported type {type(other)}")
        return self.migration_id == other.migration_id
